#include "Python.h"
#include "Numeric/arrayobject.h"
#include "math.h"
#include <stdio.h>
const double pi = 3.14159265358979;


// Shared code snippets
#include "util_ext.h"


//double max(double a, double b) { 
// 	double z; 
// 	z=(a>b)?a:b; 
// 	return z;} 
	
//double min(double a, double b) { 
// 	double z; 
// 	z=(a<b)?a:b; 
// 	return z;} 





//Innermost flux function (using w=z+h)
int _flux_function_wellbalanced(double *q_left, double *q_right,
		double normals, double g, double epsilon,
		double *edgeflux, double *max_speed) {
		
		int i;
        double z_left, z_right;
		double flux_left[2], flux_right[2];
		double z_star, w_left, h_left, h_left_star, uh_left, soundspeed_left, u_left;
		double w_right, h_right, h_right_star, uh_right, soundspeed_right, u_right;                          //h_right,
		double s_max, s_min, denom;
		
		w_left = q_left[0];
		uh_left = q_left[1];
		uh_left = uh_left*normals;
        z_left = q_left[2];
	
		w_right = q_right[0];
		uh_right = q_right[1];
		uh_right = uh_right*normals;
        z_right = q_right[2];
	  
		if (z_left-z_right != 0.0 ) printf("z_l - z_r = %f \n",z_left-z_right);
        
		//z = (z_left+z_right)/2.0;
		z_star = max(z_left, z_right);		              											//equation(7)  
				   
		// Compute speeds in x-direction
		h_left = w_left-z_left;		                                                                //This is used for computing the edgeflux[1].
		h_left_star = max(0, w_left-z_star);														//(8)
    
		if (h_left_star < epsilon) {
			u_left = 0.0; h_left_star = 0.0;
		}
		else {
			u_left = uh_left/h_left_star;
		}
	
		h_right = w_right-z_right;                                                                                                                                                                                                     
		h_right_star = max(0, w_right-z_star);														//(9)

		if (h_right_star < epsilon) {
			u_right = 0.0; h_right_star = 0.0; 
		}
		else {
			u_right = uh_right/h_right_star;
		}
  
		//printf("u_left = %g u_right = %g \n",u_left, u_right);
		soundspeed_left = sqrt(g*h_left_star);
		soundspeed_right = sqrt(g*h_right_star);
		
		s_max = max(u_left+soundspeed_left, u_right+soundspeed_right);
		if (s_max < 0.0) s_max = 0.0;
	
		s_min = min(u_left-soundspeed_left, u_right-soundspeed_right);
		if (s_min > 0.0) s_min = 0.0;
		
		
		// Flux formulas
		flux_left[0] = u_left*h_left_star;
		flux_left[1] = u_left*u_left*h_left_star + 0.5*g*h_left_star*h_left_star;

		flux_right[0] = u_right*h_right_star;
		flux_right[1] = u_right*u_right*h_right_star + 0.5*g*h_right_star*h_right_star;

		// Flux computation
		denom = s_max-s_min;
		if (denom < epsilon) {
			for (i=0; i<2; i++) edgeflux[i] = 0.0;
			*max_speed = 0.0;
		} else {
			edgeflux[0] = s_max*flux_left[0] - s_min*flux_right[0];
			edgeflux[0] += s_max*s_min*(h_right_star-h_left_star);                   //s_max*s_min*(qr[0]-ql[0]);
			edgeflux[0] /= denom;
			edgeflux[1] = s_max*flux_left[1] - s_min*flux_right[1];
			edgeflux[1] += s_max*s_min*(flux_right[0]-flux_left[0]);                 //s_max*s_min*(qr[1]-ql[1]);
			edgeflux[1] /= denom;
			edgeflux[1] *= normals;
    
		//edgeflux[1] = edgeflux[1] + 0.5*g*0.5*(h_left+h_right)*0.5*(h_left+h_right) - 0.5*g*h_left_star*h_left_star;
		//edgeflux[1] = edgeflux[1] + 0.5*g*h_right*h_right - 0.5*g*h_right_star*h_right_star;
		edgeflux[1] = edgeflux[1] + 0.5*g*h_left*h_left - 0.5*g*h_left_star*h_left_star;           //what about the h_right and h_right_star??????????????????????
		// Maximal wavespeed
        *max_speed = max(fabs(s_max), fabs(s_min));
		}  
		//printf("max_speed = %g \n",*max_speed);
    return 0;		
	}
		
		
	
	
// Computational function for flux computation
double _compute_fluxes_ext_wellbalanced(double timestep,
		double epsilon,
		double g,
		long* neighbours,
		long* neighbour_vertices,
		double* normals,
		double* areas,
		double* stage_edge_values,
		double* xmom_edge_values,
		double* bed_edge_values,
		double* stage_boundary_values,
		double* xmom_boundary_values,
		double* stage_explicit_update,
		double* xmom_explicit_update,
		int number_of_elements,
		double* max_speed_array) {
		
		double flux[2], ql[3], qr[3], edgeflux[2];
		double max_speed, normal;
		int k, i, ki, n, m, nm=0;
		
		for (k=0; k<number_of_elements; k++) {
			flux[0] = 0.0;
			flux[1] = 0.0;
			
			for (i=0; i<2; i++) {
				ki = k*2+i;
				
				ql[0] = stage_edge_values[ki];
				ql[1] = xmom_edge_values[ki];
				ql[2] = bed_edge_values[ki];
				//ql[3] = velocity_edge_values[ki];
                //ql[4] = height_edge_values[ki];
                
				n = neighbours[ki];
				if (n<0) {
					m = -n-1;
					qr[0] = stage_boundary_values[m];
					qr[1] = xmom_boundary_values[m];
					qr[2] = ql[2];
                    
				} else {
					m = neighbour_vertices[ki];
					nm = n*2+m;
					qr[0] = stage_edge_values[nm];
					qr[1] = xmom_edge_values[nm];
					qr[2] = bed_edge_values[nm];				
				}
				
				normal = normals[ki];
				_flux_function_wellbalanced(ql, qr, normal, g, epsilon, edgeflux, &max_speed);
				flux[0] -= edgeflux[0];
				flux[1] -= edgeflux[1];
				
				// Update timestep based on edge i and possibly neighbour n
				if (max_speed > epsilon) {
	                // Original CFL calculation
					
					timestep = min(timestep, 0.5*areas[k]/max_speed); //Here, CFL=1.0 is assumed. ?????????????????????????????????????????????
					if (n>=0) {
						timestep = min(timestep, 0.5*areas[n]/max_speed); //Here, CFL=1.0 is assumed. ?????????????????????????????????????????????
					}
				}
            } // End edge i (and neighbour n)
			flux[0] /= areas[k];
			stage_explicit_update[k] = flux[0];
			flux[1] /= areas[k];
			xmom_explicit_update[k] = flux[1];
			
			//Keep track of maximal speeds
			max_speed_array[k]=max_speed;
		}
		//printf("timestep = %f \n",timestep);
		return timestep;	
	}
	
	
	



//=========================================================================
// Python Glue
//=========================================================================
PyObject *compute_fluxes_ext_wellbalanced(PyObject *self, PyObject *args) {
  
    PyArrayObject 
	*neighbours, 
	*neighbour_vertices,
	*normals, 
	*areas,
	*stage_edge_values,
	*xmom_edge_values,
	*bed_edge_values,
	*stage_boundary_values,
	*xmom_boundary_values,
	*stage_explicit_update,
	*xmom_explicit_update,
	*max_speed_array;
    
  double timestep, epsilon, g;
  int number_of_elements;
    
  // Convert Python arguments to C
  if (!PyArg_ParseTuple(args, "dddOOOOOOOOOOOiO",
			&timestep,
			&epsilon,
			&g,
			&neighbours,
			&neighbour_vertices,
			&normals,
			&areas,
			&stage_edge_values,
			&xmom_edge_values,
			&bed_edge_values,
			&stage_boundary_values,
			&xmom_boundary_values,
			&stage_explicit_update,
			&xmom_explicit_update,
			&number_of_elements,
			&max_speed_array)) {
    PyErr_SetString(PyExc_RuntimeError, "comp_flux_ext.c: compute_fluxes_ext could not parse input");
    return NULL;
  }

 
  // Call underlying flux computation routine and update 
  // the explicit update arrays 
  timestep = _compute_fluxes_ext_wellbalanced(timestep,
				 epsilon,
				 g,
				 (long*) neighbours -> data,
				 (long*) neighbour_vertices -> data,
				 (double*) normals -> data,
				 (double*) areas -> data,
				 (double*) stage_edge_values -> data,
				 (double*) xmom_edge_values -> data,
				 (double*) bed_edge_values -> data,
				 (double*) stage_boundary_values -> data,
				 (double*) xmom_boundary_values -> data,
				 (double*) stage_explicit_update -> data,
				 (double*) xmom_explicit_update -> data,
				 number_of_elements,
				 (double*) max_speed_array -> data);

  // Return updated flux timestep
  return Py_BuildValue("d", timestep);
}





//-------------------------------
// Method table for python module
//-------------------------------

static struct PyMethodDef MethodTable[] = {
  {"compute_fluxes_ext_wellbalanced", compute_fluxes_ext_wellbalanced, METH_VARARGS, "Print out"},
  {NULL, NULL}
};

// Module initialisation
void initcomp_flux_ext_wellbalanced(void){
  Py_InitModule("comp_flux_ext_wellbalanced", MethodTable);
  import_array();
}
	
	
